/*
 * vim:tw=76:ts=2:sts=2:sw=2:cindent:expandtab
 */

#include <string.h>
#include <stdint.h>

#include "aes.h"
#include "eme-32-aes.h"

static inline void eme_mul_by_two(uint8_t *output, uint8_t *input)
{
  uint64_t tmp1, tmp2;
  uint64_t flagadd;

  memcpy(&tmp1, &input[8], 8);
  memcpy(&tmp2, &input[0], 8);
  flagadd = (tmp2>>63U);
  tmp2 <<= 1U;
  if (tmp1>>63U) tmp2 ^= 135;
  tmp1 <<= 1U;
  tmp1 += flagadd;

  memcpy(&output[8], &tmp1, 8);
  memcpy(&output[0], &tmp2, 8);
}

static inline void eme_xor_blocks(void *output, void *input1, void *input2)
{
  uint64_t *o = output;
  uint64_t *i1 = input1;
  uint64_t *i2 = input2;

  o[0] = i1[0] ^ i2[0];
  o[1] = i1[1] ^ i2[1];
}

void eme_aes_setup(eme_ctx *ctx, uint8_t *key, size_t keylen)
{
  uint8_t zero[16] = {0};

  aes_setup(key, keylen, &ctx->K);
  aes_ecb_encrypt(zero, zero, &ctx->K);
  eme_mul_by_two(ctx->L, zero);
}

void eme_aes_process(uint8_t C[512], uint8_t P[512], uint8_t T[16], eme_ctx *ctx, aes_func func)
{
  int j;
  uint8_t L[16] __attribute__((aligned(16)));
  uint8_t M[16] __attribute__((aligned(16)));
  uint8_t MP[16] __attribute__((aligned(16)));
  uint8_t MC[16] __attribute__((aligned(16)));

  memcpy(L, ctx->L, 16);

  for (j = 0; j < 32; j++) {
    eme_xor_blocks(&C[j*16], &P[j*16], L);
    func(&C[j*16], &C[j*16], &ctx->K);
    eme_mul_by_two(L, L);
  }

  eme_xor_blocks(MP, C, T);
  for (j = 1; j < 32; j++) {
    eme_xor_blocks(MP, MP, &C[j*16]);
  }
  func(MP, MC, &ctx->K);
  eme_xor_blocks(M, MP, MC);

  for (j = 1; j < 32; j++) {
    eme_mul_by_two(M, M);
    eme_xor_blocks(&C[j*16], &C[j*16], M);
  }
  eme_xor_blocks(C, MC, T);
  for (j = 1; j < 32; j++) {
    eme_xor_blocks(C, C, &C[j*16]);
  }

  memcpy(L, ctx->L, 16);
  for (j = 0; j < 32; j++) {
    func(&C[j*16], &C[j*16], &ctx->K);
    eme_xor_blocks(&C[j*16], &C[j*16], L);
    eme_mul_by_two(L, L);
  }
}

